#include <algorithm>

// the devil library
#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

#include "Camera.h"
#include "Ray.h"
#include "KdTree.h"
#include "Mesh.h"
#include "Exporter.h"




void
InitDevIl()
{
  ilInit();
  ILuint ImageName;
  ilGenImages(1, &ImageName);
  ilBindImage(ImageName);
  ilEnable(IL_FILE_OVERWRITE);

  //  ilRegisterFormat(IL_RGBA);
  //  ilRegisterType(IL_FLOAT);

  //  ilEnable(IL_ORIGIN_SET);
  //  ilOriginFunc(IL_ORIGIN_UPPER_LEFT);
}

bool
Camera::SnapImage(string filename,
		  KdTree *tree
		  )
{
  int x;
  int y;

  bool exportRays = true;

  
  InitDevIl();
  int components = 4;
  float *buffer = new float[components*mWidth*mHeight];
  float *pbuffer = buffer;
  //  - components*mWidth;
  
  vector<Ray> rays;

  long t1 = GetTime();

  Ray ray;

  for (y = 0; y < mHeight; y++) {
    cout<<"+";
    for (x = 0; x < mWidth; x++) {
      SetupRay(ray, x, mHeight - (y + 1));
      if (tree->CastRay(ray))
	if (ray.intersections.size()) {
	  sort(ray.intersections.begin(), ray.intersections.end());
	  MeshInstance *mesh = (MeshInstance*)ray.intersections[0].mObject;
	  RgbColor color(1,1,1);
	  if (mesh->GetMesh()->mMaterial)
	    color = mesh->GetMesh()->mMaterial->mDiffuseColor;

	  pbuffer[0] = color.r;
	  pbuffer[1] = color.g;
	  pbuffer[2] = color.b;
	  pbuffer[3] = 1.0f;

	}
      pbuffer+=components;
      
      if (exportRays && (x==222) && (y==97))
	rays.push_back(ray);
    }
    //    pbuffer-=2*components*mWidth;
  }
  
  long t2 = GetTime();
  cout<<"#RAY_CAST_TIME\n";
  cout<<TimeDiff(t1, t2)<<"\n";

  
  cout<<"Saving image"<<endl;

  ilRegisterType(IL_FLOAT);
  ilTexImage(mWidth, mHeight, 1, 4, IL_RGBA, IL_FLOAT, buffer);
  ilSaveImage((char *const)filename.c_str());
  delete buffer;
  
  Exporter *exporter = NULL;
  if (exportRays) {
    exporter = Exporter::GetExporter(filename + "-rays" + ".x3d");
    exporter->SetWireframe();
    exporter->ExportKdTree(*tree);
	//exporter->ExportBspTree(*bsptree);
    exporter->ExportRays(rays, 10000);
    exporter->SetFilled();
    int k =0;
    for (int j=0; j < rays.size(); j++)
      if (rays[j].kdLeaves.size()) {
	Ray *ray = &(rays[j]);
	int i;
	if (1)
	for (i= 0; i < ray->kdLeaves.size(); i++)
	  exporter->ExportBox(tree->GetBox(ray->kdLeaves[i]));
	if (0)
	for (i= 0; i < ray->meshes.size(); i++)
	  exporter->ExportIntersectable(ray->meshes[i]);

      }

    delete exporter;
  }
  
  
  return true;
}


void
Camera::SetupRay(Ray &ray, const int x, const int y)
{
  Vector3 xv = mRight*((x - mWidth/2)/(float)mWidth);
  Vector3 yv = mUp*((y - mHeight/2)/(float)mHeight);
  Vector3 target = xv + yv + mDirection;

  ray.intersections.clear();
  ray.kdLeaves.clear();
  ray.meshes.clear();
  ray.bspLeaves.clear();

  ray.Init(mPosition, target, Ray::LOCAL_RAY);
  
}
