#ifndef _SamplingPreprocessor_H__
#define _SamplingPreprocessor_H__

#include <fstream>
using namespace std;

#include "Preprocessor.h"

struct SimpleRay {
  Vector3 mOrigin;
  Vector3 mDirection;
};

/** Sampling based visibility preprocessing. The implementation is based on heuristical
    sampling of view space */
class SamplingPreprocessor : public Preprocessor {
public:
	int mPass;

  int mSamplesPerPass;
  int mTotalSamples;
  int mKdPvsDepth;
  ofstream mStats;
  ObjectContainer mObjects;
  RayContainer mSampleRays;
  int mBspConstructionSamples;

	
  SamplingPreprocessor();

  ~SamplingPreprocessor();

  virtual bool ComputeVisibility();

  void
  SetupRay(Ray &ray, 
		   const Vector3 &point, 
		   const Vector3 &direction,
		   const int type);

  KdNode *
  GetNodeForPvs(KdLeaf *leaf);

  int
  AddNodeSamples(Intersectable *object, const Ray &ray);

  void
  HoleSamplingPass();

	int
	CastRay(Intersectable *object, Ray &ray, const bool reverseSample);

	KdNode *
	GetNodeToSample(Intersectable *object);
	void
	VerifyVisibility(Intersectable *object);

	int
	CastEdgeSamples(
									Intersectable *object,
									const Vector3 &point,
									MeshInstance *mi,
									const int samples
									);

	/** Processes the BSP based view cells during a pass.
		@param collectSamplesForBps if still in sample collection phase
		@param ray the current ray
		@param object the currently processed object
		@param faceIndex the sampled face of the object
		@param contributingSamples samples contributing to pvs
		@param sampleContributions contribution of the samples

		@returns true if BSP tree has been constructed, false if not yet

	*/
	bool ProcessBspViewCells(bool collectSamplesForBsp,
							 Ray &ray,
							 Intersectable *object,
							 int faceIndex,
							 int &contributingSamples,
							 int &sampleContributions);

	/** Adds objects samples to bsp view cells.
	*/
	int AddObjectSamples(Intersectable *obj, const Ray &ray);

	bool BuildBspTree();

	void ExportSplits(const ObjectContainer &objects);

	void ExportBspPvs(const ObjectContainer &objects);

	/** Post processes view cells (i.e., merges or subdivides view cells based
		on the PVS and the ray sets.
		@returns number of merged view cells.
	*/
	int PostprocessViewCells(const RayContainer &rays);

};



#endif
